/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * Portions Copyrighted 2009 Jaroslav Tulach
 */

package cz.xelfi.quoridor.freemarkerdor;

import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.UniformInterfaceException;
import com.sun.jersey.api.client.WebResource;
import com.sun.jersey.api.container.httpserver.HttpServerFactory;
import com.sun.jersey.api.core.PackagesResourceConfig;
import com.sun.jersey.api.core.ResourceConfig;
import com.sun.jersey.api.view.Viewable;
import com.sun.net.httpserver.HttpServer;
import cz.xelfi.quoridor.Board;
import cz.xelfi.quoridor.IllegalPositionException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.concurrent.Callable;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.FormParam;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Cookie;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.NewCookie;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import org.openide.util.Exceptions;
import org.w3c.dom.Document;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
@Path("/")
public final class UI {
    private static final String version;
    static {
        Properties p = new Properties();
        try {
            InputStream is = FreemarkerProcessor.class.getResourceAsStream("/META-INF/maven/org.apidesign/freemarkerdor/pom.properties"); // NOI18N
            if (is != null) {
                p.load(is);
            }
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        version = p.getProperty("version", "unknown"); // NOI18N
    }
    private static WebResource base;

    @Context
    private HttpHeaders headers;
    private String user;
    private String uuid;

    public UI() {
    }

    private Viewable checkLogin() {
        Cookie cookie = headers.getCookies().get("login");
        if (cookie != null) {
            String id = cookie.getValue();
            String us;
            try {
                us = base.path("login").queryParam("id", id).
                    accept(MediaType.TEXT_PLAIN).get(String.class);
            } catch (UniformInterfaceException ex) {
                ex.printStackTrace();
                us = "";
            }
            if (us.length() > 0) {
                user = us;
                uuid = id;
                return null;
            }
        }
        return viewable("login.fmt", null);
    }

    @POST
    @Path("login")
    @Produces(MediaType.TEXT_HTML)
    public Response login(
        @FormParam("name") String name, @FormParam("password") String password
    ) throws Exception {
        uuid = base.path("login").queryParam("name", name).queryParam("password", password).
            accept(MediaType.TEXT_PLAIN).put(String.class);
        if (uuid != null) {
            user = name;
            NewCookie nc = new NewCookie("login", uuid, null, null, null, 3600 * 24 * 7, false);
            return Response.ok().cookie(nc).entity(viewable("login.fmt", null)).build();
        } else {
            Viewable v = viewable("login.fmt", null, "message", "Invalid name or password: " + name);
            return Response.status(1).entity(v).build();
        }
    }

    @GET
    @Produces(MediaType.TEXT_HTML)
    public Viewable welcome() {
        Viewable v = checkLogin();
        if (v != null) {
            return v;
        }
        return welcomeImpl();
    }

    private Response board(String id) {
        return board(id, "", null);
    }
    @GET
    @Path("games/{id}/")
    @Produces(MediaType.TEXT_HTML)
    public Response board(
        @PathParam("id") String id,
        @QueryParam("format") @DefaultValue("") String format,
        @QueryParam("move") @DefaultValue("-1") String move
    ) {
        return board(id, null, format, move);
    }
    private Response board(@PathParam("id") String id, String msg, String format, String m) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.ok().entity(v).build();
        }
        int move;
        try {
            move = Integer.parseInt(m);
        } catch (NumberFormatException ex) {
            move = -1;
        }
        WebResource url = base.path("games").path(id);
        if (move >= 0) {
            url = url.queryParam("move", "" + move);
        }
        ResponseBuilder resp = Response.ok();
        Cookie cFormat = headers.getCookies().get("format");
        if (format.length() == 0) {
            if (cFormat != null) {
                format = cFormat.getValue();
            }
        } else {
            if (cFormat == null || !format.equals(cFormat.getValue())) {
                resp.cookie(new NewCookie("format", format));
            }
        }

        Document doc = url.accept(MediaType.TEXT_XML).get(Document.class);
        Board b;
        try {
            b = Board.valueOf(doc.getElementsByTagName("board").item(0).getTextContent());
        } catch (IllegalPositionException ex) {
            Exceptions.printStackTrace(ex);
            b = Board.empty();
        }
        v = viewable("game.fmt", doc, "message", msg, "format", format, "board", b);
        return resp.entity(v).build();
    }

    @GET
    @Path("games/{id}/move")
    @Produces(MediaType.TEXT_HTML)
    public Response move(
        @PathParam("id") String id,
        @QueryParam("type") String type,
        @QueryParam("direction") String direction,
        @QueryParam("direction-next") @DefaultValue("") String directionNext,
        @QueryParam("column") @DefaultValue("") String column,
        @QueryParam("row") @DefaultValue("") String row
    ) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.ok().entity(v).build();
        }
        WebResource wr = base.path("games").path(id).
            queryParam("loginID", uuid).
            queryParam("player", user);
        try {
            if (type.equals("resign")) {
                wr.queryParam("move", "RESIGN").put();
                return board(id);
            }
            if (type.equals("fence")) {
                wr.queryParam("move", direction.charAt(0) + column + row).put();
                return board(id);
            }
            if (type.equals("move")) {
                wr.queryParam("move", direction + directionNext).put();
                return board(id);
            }
        } catch (UniformInterfaceException ex) {
            return board(id, "WRONG_MOVE", "-1");
        }
        return board(id);
    }

    @GET
    @Path("games/create")
    @Produces(MediaType.TEXT_HTML)
    public Response create(
        @QueryParam("white") String white,
        @QueryParam("black") String black
    ) {
        Viewable v = checkLogin();
        if (v != null) {
            return Response.status(Response.Status.FORBIDDEN).entity(v).build();
        }

        if (user.equals(white) || user.equals(black)) {
            Object obj =
                base.path("games").
                queryParam("loginID", uuid).
                queryParam("white", white).
                queryParam("black", black).accept(MediaType.TEXT_XML).post(Document.class);
            return Response.ok(welcomeImpl()).build();
        } else {
            return Response.status(Response.Status.NOT_FOUND).
                entity(welcomeImpl("message", "You (" + user + ") must be white or black!")).build();
        }
    }

    private Viewable welcomeImpl(Object... args) {
        final Document got = base.path("games").accept(MediaType.TEXT_XML).get(Document.class);
        return viewable("index.fmt", got, args);
    }

    //
    // start the server
    //

    public static void main(String[] args) throws Exception {
        int port = 9998;
        if (args.length > 1) {
            port = Integer.parseInt(args[0]);
        }


        Callable<Void> r = startServers(port);

        if (args.length < 2 || !args[1].equals("--kill")) {
            System.out.println("Hit enter to stop it...");
            System.in.read();
        } else {
            synchronized (UI.class) {
                UI.class.wait();
            }
        }
        r.call();
        System.exit(0);
    }

    static Callable<Void> startServers(int port) throws Exception {

        if (System.getProperty("quoridor.dir") == null) {
            File home = new File(System.getProperty("user.home"));
            File quoridor = new File(home, ".quoridor");
            System.setProperty("quoridor.dir", quoridor.getPath());
        }

        ResourceConfig rc = new PackagesResourceConfig(
            "cz.xelfi.quoridor.webidor",
            "cz.xelfi.quoridor.freemarkerdor"
        );

        Client client = new Client();
        base = client.resource(new URI("http://localhost:" + port + "/api/"));

        final String baseUri = "http://localhost:" + port + "/";
        final HttpServer server = HttpServerFactory.create(baseUri, rc);
        server.start();
        System.out.println("Quoridor started at port " + port);

        return new Callable<Void>() {
            public Void call() throws Exception {
                server.stop(0);
                return null;
            }
        };
    }

    private Viewable viewable(String page, Document doc, Object... more) {
        ResourceBundle rb = null;
        for (Locale l : headers.getAcceptableLanguages()) {
            try {
                rb = ResourceBundle.getBundle("cz.xelfi.quoridor.freemarkerdor.UI.Bundle", l);
            } catch (MissingResourceException e) {
                // OK
            }
        }
        if (rb == null) {
            rb = ResourceBundle.getBundle("cz.xelfi.quoridor.freemarkerdor.UI.Bundle", Locale.ENGLISH);
        }

        Map<String,Object> map = new HashMap<String,Object>();
        map.put("doc", doc);
        map.put("user", user);
        map.put("bundle", rb);
        map.put("now", System.currentTimeMillis());
        map.put("version", version);
        for (int i = 0; i < more.length; i += 2) {
            map.put((String)more[i],more[i + 1]);
        }
        return new Viewable(page, map);
    }

}
